//////////////////////////////////////////////////////////////////////////
// TSXtruder 1.0 by Bendy's 3D Systems
// Copyright (c) 1999
//
// Based on Extruder 1.1 by Bryce Denney
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// The complete GNU public license can be found in this directory.
//
//////////////////////////////////////////////////////////////////////////
//
// hfield.cc
//
// Function definitions for the height field class (Hfield).
// It is described in hfield.h.
//
//////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "resource.h"
#include "tga.h"
#include "TSXDialog.h"
#include "hfield.h"

#include "tsx.h"

Hfield::Hfield(
		int _hres, int _vres,       // grid resolution (horiz and vert)
		float _base_size,           // height of the rectangular base
		float max_height,           // maximum height of extruded part
		TGAfile *tga,               // TGAfile object containing the 2D image
		float _x, float _y,			// Object size
		int smooth,	double sfac,	// Smooth? and factor
		int face_t,					// triangulate?
		char *file,					// Name of the file. used for texture
		int texture,				// Apply texture
		int bump)					// Apply bump
{
	int	xi,yi;
	int	tga_w=tga->get_width();
	int	tga_h=tga->get_height();

	hres = _hres; vres = _vres;
	base_size = _base_size;
	data = new Hfield_pixel[hres*vres];
	assert(data);
	int vnum=0;  // the vertex number (0 to hres*vres-1)
	float hfactor = 2 / (float)(hres-1);
	float vfactor = 2 / (float)(vres-1);
	float height_factor = max_height / 256;
	float xfactor = (float)(tga_w-1) / (float)(hres-1);
	float yfactor = (float)(tga_h-1) / (float)(vres-1);
	//Pixel pixel;

	// Make heightfield vertices
		// Object size is (-1,-1),(1,1)

	// Double check that smoothing is off for res >
	// half of the image size. This is also controlled
	// by the dialog box.
	if (hres>tga_w/2 || vres>tga_h/2)
		smooth=0;
	
	for (int v=0; v<vres; v++)
	for (int h=0; h<hres; h++)
	{
		// assign x and y values between -1 and 1.
		data[vnum].x = -1 + hfactor * h;
		data[vnum].y = -1 + vfactor * v;
		// find the corresponding point in the TGA file
		xi = (int)((float) h * xfactor);
        yi = (int)((float) v * yfactor);

		if (smooth)
			data[vnum].z = smoothPixel(tga,xi,yi,1/sfac);
		else
			data[vnum].z = (float)tga->get_grey_value(xi, yi);
		data[vnum].z = data[vnum].z * height_factor;
		// scale it so that 0 --> 0 and 255 --> height_factor.
		data[vnum].vertex=vnum;    // assign the vertex number
		vnum++;
	}
	// reassure ourselves that the vertex numbers were 0 to hres*vres-1
	assert(vnum==hres*vres);
	
	// Calculate size of object
	xsize = _x;
	ysize = _y;
	
	// Triangulation?
	face_type=face_t;

	if (filename!=NULL)
	{
		filename = new char [strlen(file)];
		assert(filename);
		strcpy(filename,file);
	}
	add_texture=texture;
	add_bumpmap=bump;
}

// Returns the smoothed pixel value at pos
// x,y. Smoothing as taken over an area relative
// to hres,vres. Smoothing is such that the 
// furthest pixels have weight of 1/param. 
// Pixels between are weighted gaussianly.
float
Hfield::smoothPixel(TGAfile *tga, int x, int y, double param)
{
	int		xs=tga->get_width(),ys=tga->get_height();
	int		xdelta = int(xs / (2.0*(hres)));
	int		ydelta = int(ys / (2.0*(vres)));
	float	norm=0, fat;
	float	z=0,weight;
	int		yi,xi;
	int		n_times=0,n_times2=0;

	// Controls fatness of gaussian.
	//  weights of furthest pixels are 1/param
	fat = float((xdelta*xdelta+ydelta*ydelta)/log(param));

	yi = y-ydelta;

	for (;yi<=y+ydelta;yi++)
	{
		for (xi=x-xdelta;xi<=x+xdelta;xi++)
		{
			n_times2++;
			if (xi>=0 && yi>=0 && xi<xs && yi<ys)
			{
				weight=float(exp(-(pow(xi-x,2)+pow(yi-y,2))/fat));
				z += tga->get_grey_value(xi, yi)*weight;
				norm+=weight;
				n_times++;
			}
		}
	}
	assert(norm);
	if (norm)
		return z/norm;
	else
		return z;
}

// All code below here added by Brendan Hack (bendy@bendys.com)
// Began 9/11/97

int
Hfield::calcNSemi(void)
{
	int h,v,n_faces=0,vnum;

	for (h=0;h<hres-1;h++)
		for(v=0;v<vres-1;v++)
		{
			vnum=h+v*hres;
			if (data[vnum].z ==
					data[vnum+1].z &&
					data[vnum+hres].z ==
					data[vnum+hres+1].z &&
					data[vnum].z ==
					data[vnum+hres].z)
						n_faces++;
			else
						n_faces+=2;
		}
	return n_faces;

}

// Function to make the heightfield as a new object in Truespace 3.
// Used instead of writing the object to file.
int
Hfield::insertObject(CTSXDialog &dlg)
{
	tsxPOLYHEDRON	*poly;			// Pointer to final poly
	CtsxVector3f	*verts;			// Array of polys vertices
	CtsxFaceVx		*face_verts;	// Array of vertices for face
	CtsxFace		**faces;		// Array of pointers to faces in poly
	CtsxFace		*face=NULL;		// A Face
	CtsxUV			*uv;			// UV map
	tsxMATERIAL		*texture;		// texture for object

	int				n_faces,n_verts,vnum,h,v,i;
	int				base_idx;		// Idx into vertex array of start of
									// base vertices;
	int	progress=0, prog_next, prog_inc;	// Progress meter;

	// Make the poly
	if (tsxPolyhCreate(&poly)<0)
		assert(FALSE);

	// Make vertices and UV space
	n_verts = 4+hres*vres;
	// Allocate memory for vertices and UV space
	verts = (CtsxVector3f*)tsxCalloc(n_verts,sizeof(CtsxVector3f));
	assert(verts);
	uv = (CtsxUV*)tsxCalloc(n_verts,sizeof(CtsxUV));
	assert(uv);

	// Set heightfield vertices and uv space
	// NEW FOR BETA 2. Scale to aspect of image
	for (vnum=0;vnum<vres*hres;vnum++)
	{
		verts[vnum].x = data[vnum].x*xsize;
		verts[vnum].y = data[vnum].y*ysize;
		verts[vnum].z = data[vnum].z;
		uv[vnum].u = float(0.5) + (data[vnum].x/2);
		uv[vnum].v = float(0.5) + (data[vnum].y/2);
	}
	// Set base vertices and uv space
	// (clockwise order looking from below)
	// Z axis is pointing into of the screen
	// base_idx is index into first base vertex
	//
	//                Y
	//                ^
	//                |
	// (bi+3) nw o---------o ne (base_idx)
	//           |    |    |
	//      X <--|----+    |
	//           |         |  
	// (bi+2) sw o---------o se  (bi+1)
	//
	//
	base_idx = vnum;
	verts[vnum].x=-xsize; verts[vnum].y=ysize; verts[vnum].z=-base_size; //NE
	uv[vnum].u=0;uv[vnum].v=1;vnum++;
	verts[vnum].x=-xsize;verts[vnum].y=-ysize;verts[vnum].z=-base_size;  //SE
	uv[vnum].u=0;uv[vnum].v=0;vnum++; 
	verts[vnum].x=xsize;verts[vnum].y=-ysize; verts[vnum].z=-base_size;//SW
	uv[vnum].u=1;uv[vnum].v=0;vnum++; 
	verts[vnum].x=xsize; verts[vnum].y=ysize;verts[vnum].z=-base_size; //NW
	uv[vnum].u=1;uv[vnum].v=1;vnum++; 
	assert(vnum==n_verts);
	// Allocate vertices and UV to poly
	tsxPolyhSetVxs(poly,n_verts,verts);
	tsxPolyhSetUVArray(poly,n_verts,uv);

	// Make the faces
	// Precalculate # of faces
	switch(face_type)
	{
		case(F_PLANE):n_faces = 5+(hres-1)*(vres-1);
				if (!isPlanar()) // Warn if non-planar faces.
					AfxMessageBox(IDS_NONPLANAR);break;
		case(F_TRIANGLE):n_faces = 5+2*(hres-1)*(vres-1);break;
		case(F_SEMI):n_faces=5;n_faces+=calcNSemi();break;
	}
	
	// Allocate memory for faces
	faces = (CtsxFace**)tsxCalloc(n_faces,sizeof(CtsxFace*));
	assert(faces);
	i = 0;
	prog_next = prog_inc = n_faces/PROGRESSBARINC;
	// Set face info
	// Do top faces

	for (h=0;h<hres-1;h++)
		for(v=0;v<vres-1;v++)
		{
			vnum = h+v*hres;	// Vertex Number
			switch(face_type)
			{
				case(F_PLANE): makePlaneFace(vnum,verts,i,faces);break;
				case(F_TRIANGLE): makeTriangularFaces(vnum,verts,i,faces);break;
				case(F_SEMI): makeSemiFaces(vnum,verts,i,faces);break;
			}
			if (i>prog_next)
			{
				progress +=100/PROGRESSBARINC;
				prog_next+=prog_inc;
				dlg.SetProgress(progress);
			}
		}

	// Do Bottom Face
	face_verts = (CtsxFaceVx*)tsxCalloc(4,sizeof(CtsxFaceVx)); // faces have 4 verts
	assert(face_verts);
	// Set fields in verts: vix, uvix
	// Clockwise order from below.
	for (h=0;h<4;h++)
		setFaceVert(face_verts[h],base_idx+h);
	face=makeFace(face,face_verts,4);
	// Add to face list
	faces[i++]=face;

	// Left Face
	face_verts = (CtsxFaceVx*)tsxCalloc(2+vres,sizeof(CtsxFaceVx));
	assert(face_verts);
	// Bottom verts first
	setFaceVert(face_verts[0],base_idx+SE);
	setFaceVert(face_verts[1],base_idx+NE);
	// Now the rest of them in clockwise order
	
	for (vnum=2,v=vres-1;v>=0;v--,vnum++)
		setFaceVert(face_verts[vnum],hres*v);
	// Vertex list is done. Now setup the face
	face=makeFace(face,face_verts,2+vres);
	// Add to face list
	faces[i++]=face;

	// Right Face
	face_verts = (CtsxFaceVx*)tsxCalloc(2+vres,sizeof(CtsxFaceVx));
	assert(face_verts);
	// Bottom verts first
	setFaceVert(face_verts[0],base_idx+NW);	
	setFaceVert(face_verts[1],base_idx+SW);
	// Now the rest of them in clockwise order
	for (vnum=2,v=0;v<vres;v++,vnum++)	
		setFaceVert(face_verts[vnum],hres*v+(hres-1));
	// Vertex list is done. Now setup the face
	face=makeFace(face,face_verts,2+vres);
	// Add to face list
	faces[i++]=face;
	
	// "Top" Face
	face_verts = (CtsxFaceVx*)tsxCalloc(2+hres,sizeof(CtsxFaceVx));
	assert(face_verts);
	// Bottom verts first
	setFaceVert(face_verts[0],base_idx+NE);
	setFaceVert(face_verts[1],base_idx+NW);
	// Now the rest of them in clockwise order
	for (vnum=2,h=hres-1;h>=0;h--,vnum++)	
		setFaceVert(face_verts[vnum],h+(vres-1)*hres);
	// Vertex list is done. Now setup the face
	face=makeFace(face,face_verts,2+hres);
	// Add to face list
	faces[i++]=face;
	
	// "Bottom" Face
	face_verts = (CtsxFaceVx*)tsxCalloc(2+hres,sizeof(CtsxFaceVx));
	assert(face_verts);
	// Bottom verts first
	setFaceVert(face_verts[0],base_idx+SW);
	setFaceVert(face_verts[1],base_idx+SE);
	// Now the rest of them in clockwise order
	for (vnum=2,h=0;h<hres;h++,vnum++)	
		setFaceVert(face_verts[vnum],h);
	// Vertex list is done. Now setup the face
	face=makeFace(face,face_verts,2+hres);
	// Add to face list
	faces[i++]=face;


	if (i!=n_faces)
	{
		CString str;
		char	it[32],nft[32];
		AfxFormatString2(str,IDS_BADFACES,_itoa(i,it,10),_itoa(n_faces,nft,10));
		AfxMessageBox(str);
		tsxFree(verts);
		tsxFree(uv);
		tsxFree(faces);
		tsxFree(poly);
		return 0;
	}
	
	dlg.SetProgress(0);
	// Now finish off making the poly
	tsxPolyhSetFaces(poly,n_faces,faces);
	dlg.SetProgress(10);
	if (tsxPolyhComputeFaceAdjacency(poly)<0)
		assert(FALSE);
	dlg.SetProgress(20);
	tsxPolyhMarkTopologyChange(poly);
	dlg.SetProgress(30);
	tsxPolyhMarkGeomChange(poly);
	dlg.SetProgress(40);
	if (tsxPolyhInitializeMaterials(poly)<0)
		assert(FALSE);
	dlg.SetProgress(50);
	// NEW FOR BETA 2
	// add texture
	if (add_texture)
	{
		CtsxTextureProps	tprops;
		CtsxSurfaceProps	sprops;

		tprops.offsetu=tprops.offsetv=0;
		tprops.repeatsu=tprops.repeatsv=1;
		tprops.flags=tsxTXTR_ON;

		sprops.ka=sprops.ks=float(0.1);
		sprops.exp=0;

		// make default material
		if (tsxMaterialCreate(&texture)<0)
			assert(FALSE);
		dlg.SetProgress(60);
		// make it a texture map and set the filename
		tsxMaterialSetTextureType(texture,tsxTEXTURE_NONE);
		tsxMaterialSetSurfaceProps(texture,&sprops);
		//tsxMaterialSetProcTexture(texture,tsxTEXTURE_NONE);
		tsxMaterialSetTextureFilename(texture,filename);
		tsxMaterialSetTextureProps(texture,&tprops);
		tsxMaterialSetShadingType(texture,tsxSHADER_PHONG);
		tsxMaterialSetFacetingType(texture,tsxFACET_SMOOTH);
		dlg.SetProgress(70);
		if (add_bumpmap)
		{
			CtsxBumpProps	bprops;

			bprops.offsetu=bprops.offsetv=0;
			bprops.repeatsu=bprops.repeatsv=1;
			bprops.amplitude=(float)0.2;
			bprops.flags=tsxBUMP_ON;
		
			tsxMaterialSetBumpFilename(texture,filename);
			tsxMaterialSetBumpProps(texture,&bprops);
		}
		dlg.SetProgress(80);
		tsxPolyhPaint(poly,texture);
		tsxPolyhMarkMaterialChange(poly);
	//	tsxMaterialDestroy(texture);
	}
	dlg.SetProgress(90);
	// Give It A Name		New for V1.0
	tsxGNodeSetName((tsxGNODE *) poly, "HeightField");
	// Put it in da scene
	tsxSceneAddObject((tsxGNODE*) poly, e_tsxFALSE);
	dlg.SetProgress(100);
	return n_faces;

}

// Makes a face from n_verts vertices pointed to by verts
// Assigns faves and calculates normal vectors
CtsxFace *
Hfield::makeFace(CtsxFace *face, CtsxFaceVx *verts, int n_verts)
{
	face = tsxFaceAlloc();
	assert(face);
	face->pFVxs = verts;
	face->nbrVxs = n_verts;
	// Set normal of face;
	// Due to the nature of the faces made we can just use 
	// the first, second and last vertices for the normal rays. It
	// is guaranteed they won't be colinear. If you don't believe me
	// draw the top, bottom and side poly's and check it out.

	face->normVxs[tsx_NORM_VXS_RAY1]   = 0;
	face->normVxs[tsx_NORM_VXS_ORIGIN] = 1;
	face->normVxs[tsx_NORM_VXS_RAY2]   = 2;
	return face;
}

void
Hfield::makePlaneFace(int vnum, CtsxVector3f *verts, int &i, CtsxFace **faces)
{
	CtsxFaceVx	*face_verts;
	CtsxFace	*face=NULL;

	face_verts = (CtsxFaceVx*)tsxCalloc(4,sizeof(CtsxFaceVx));
					// faces have 4 verts
	assert(face_verts);
	// Set fields in verts: vix, uvix
	setFaceVert(face_verts[0],vnum);
	setFaceVert(face_verts[1],vnum+hres);
	setFaceVert(face_verts[2],vnum+1+hres);
	setFaceVert(face_verts[3],vnum+1);

	// Vertex list is done. Now setup the face
	face=makeFace(face,face_verts,4);
	
	// Add to face list
	faces[i++]=face;
}

void
Hfield::makeTriagAFaces(int vnum, CtsxVector3f *verts, int &i, CtsxFace **faces)
{
	CtsxFaceVx	*face_verts;
	CtsxFace	*face=NULL;

	// Make 1st triangle
	face_verts = (CtsxFaceVx*)tsxCalloc(3,sizeof(CtsxFaceVx));
	assert(face_verts);
					
	setFaceVert(face_verts[0],vnum);
	setFaceVert(face_verts[1],vnum+hres+1);
	setFaceVert(face_verts[2],vnum+1);
		// Vertex list is done. Now setup the face
	face=makeFace(face,face_verts,3);
			
	// Add to face list
	faces[i++]=face;
	// Make 2nd triangle
	face_verts = (CtsxFaceVx*)tsxCalloc(3,sizeof(CtsxFaceVx));
	assert(face_verts);
	setFaceVert(face_verts[0],vnum);
	setFaceVert(face_verts[1],vnum+hres);
	setFaceVert(face_verts[2],vnum+hres+1);			
		
	// Vertex list is done. Now setup the face
	face=makeFace(face,face_verts,3);
				
	// Add to face list
	faces[i++]=face;
}

void
Hfield::makeTriagBFaces(int vnum, CtsxVector3f *verts, int &i, CtsxFace **faces)
{
	CtsxFaceVx	*face_verts;
	CtsxFace	*face=NULL;

	// Make 1st triangle
	face_verts = (CtsxFaceVx*)tsxCalloc(3,sizeof(CtsxFaceVx));
	assert(face_verts);
					
	setFaceVert(face_verts[0],vnum);
	setFaceVert(face_verts[1],vnum+hres);
	setFaceVert(face_verts[2],vnum+1);
		// Vertex list is done. Now setup the face
	face=makeFace(face,face_verts,3);
				
	// Add to face list
	faces[i++]=face;

	// Make 2nd triangle
	face_verts = (CtsxFaceVx*)tsxCalloc(3,sizeof(CtsxFaceVx));
	assert(face_verts);
	setFaceVert(face_verts[0],vnum+1);
	setFaceVert(face_verts[1],vnum+hres);
	setFaceVert(face_verts[2],vnum+hres+1);				
			
	// Vertex list is done. Now setup the face
	face=makeFace(face,face_verts,3);
				
	// Add to face list
	faces[i++]=face;
}
void
Hfield::makeTriangularFaces(int vnum, CtsxVector3f *verts, int &i, CtsxFace **faces)
{
	if (fabs((verts[vnum].z-verts[vnum+hres+1].z)) <
		fabs((verts[vnum+1].z-verts[vnum+hres].z)))
		makeTriagAFaces(vnum,verts,i,faces); 
	else
		makeTriagBFaces(vnum,verts,i,faces);
}

void
Hfield::makeSemiFaces(int vnum, CtsxVector3f *verts, int &i, CtsxFace **faces)
{
	if (verts[vnum].z ==
		verts[vnum+1].z &&
		verts[vnum+hres].z ==
		verts[vnum+hres+1].z &&
		verts[vnum].z ==
		verts[vnum+hres].z)
			makePlaneFace(vnum,verts,i,faces);
	else
			makeTriangularFaces(vnum,verts,i,faces);
}

BOOL
Hfield::isPlanar(void)
{
	int h,v,vnum;

	for (h=0;h<hres-1;h++)
		for(v=0;v<vres-1;v++)
		{
			vnum=h+v*hres;
			if (!(data[vnum].z ==
					data[vnum+1].z &&
					data[vnum+hres].z ==
					data[vnum+hres+1].z &&
					data[vnum].z ==
					data[vnum+hres].z))
				return FALSE;
		}
	return TRUE;
}

